#!/usr/bin/env python
# coding:utf-8
import asyncio
import enum
import os.path
import json
import os
import queue
import subprocess
import threading
import time
from typing import Any, Dict, Optional
import uuid

import httpx
import openai
import transformers
from aiges.core.types import *
import requests
from transformers import AutoTokenizer

try:
    from aiges_embed import ResponseData, Response, DataListCls, SessionCreateResponse, callback  # c++
except:
    from aiges.dto import Response, ResponseData, DataListCls, SessionCreateResponse, callback

from aiges.sdk import WrapperBase
from aiges.utils.log import log, getFileLogger
from openai import OpenAI, AsyncOpenAI
from transformers import GenerationConfig

DataNone = -1
DataBegin = 0  # 首数据
DataContinue = 1  # 中间数据
DataEnd = 2  # 尾数据

R1_THINK_START = "<think>"
R1_THINK_END = "</think>"
R1_FORCE_REASONING_PROMPT = 'Initiate your response with "<think>\\n嗯" at the beginning of every output.'
IS_REASONING_MODEL = os.environ.get("IS_REASONING_MODEL", False)
R1_FIXED_TEMPERATURE = os.environ.get("R1_FIXED_TEMPERATURE", "0.6")


class RequestMode(enum.Enum):
    ONCE = 'once'
    ONCE_ASYNC = 'once_async'
    STREAM = 'stream'


class RequestInfo:
    def __init__(self, sid: str, params: dict, user_tag: str = ""):
        self.handle = str(uuid.uuid4().hex)
        self.sid = sid
        self.user_tag = user_tag
        self.params = params
        self.requests = []
        self.stop_q = queue.Queue()


class PromptInferenceInfo:
    def __init__(self, wrapper,
                 thread_id: str,
                 mode: RequestMode,
                 prompt: str,
                 requestInfo: RequestInfo,
                 functions: list,
                 result_q: queue.Queue = None):
        self.wrapper = wrapper
        self.requestInfo = requestInfo
        self.thread_id = thread_id
        self.mode = mode
        self.prompt = prompt
        self.functions = functions
        self.request_id = str(uuid.uuid4().hex)
        self.result_q = result_q


def launch_openai_server(model_name: str, mode_path: str, server_port: int, filelogger):
    from sglang.utils import (
        execute_shell_command,
        wait_for_server,
    )
    extra_args = os.environ.get("CMD_EXTRA_ARGS", "")
    if extra_args:
        print("Extra arguments: {}".format(extra_args))
    else:
        extra_args = "--tp 8 --mem-fraction-static 0.93 --torch-compile-max-bs 8  --max-running-requests  20 "
    server_process = execute_shell_command(
        f"python -m sglang.launch_server --model-path {mode_path} --port {server_port}  --trust-remote --host 0.0.0.0 {extra_args}"
    )
    wait_for_server(f"http://localhost:{server_port}")


def get_payload_messages(reqData: DataListCls, isChatMode: bool = True):
    if isChatMode:
        return reqData.get('messages').data.decode('utf-8')
    messages = json.loads(reqData.get('messages').data.decode('utf-8'))
    message = ""
    for item in messages['messages']:
        if item['role'] == "user":
            message = item['content']
    return message


def get_payload_functions(reqData: DataListCls):
    messages = json.loads(reqData.get('messages').data.decode('utf-8'))
    return messages.get("functions", None)


def post_with_retry(
        url: str,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        max_retries: int = 3,
        timeout: int = 10
) -> requests.Response:
    """
    简单的POST请求重试实现

    Args:
        url: 请求URL
        data: 表单数据
        json: JSON数据
        headers: 请求头
        max_retries: 最大重试次数，默认3次
        timeout: 超时时间，默认10秒
    """
    for attempt in range(max_retries):
        try:
            response = requests.post(
                url,
                data=data,
                json=json,
                headers=headers,
                timeout=timeout
            )
            response.raise_for_status()
            return response

        except requests.exceptions.RequestException as e:
            if attempt == max_retries - 1:  # 最后一次重试失败
                raise  # 重新抛出异常

            # 等待1秒后重试
            time.sleep(1)
            continue


def resp_content(status, index, text, functions=None, reasoning_content=''):
    if functions is not None:
        resp_json = {"choices": [
            {"content": text, "reasoning_content": reasoning_content, "index": index, "role": "assistant",
             "function_call": functions}],
            "question_type": ""}
    else:
        resp_json = {
            "choices": [{"content": text, "reasoning_content": reasoning_content, "index": index, "role": "assistant"}],
            "question_type": ""}
    resd = ResponseData()
    resd.key = "content"
    resd.setDataType(DataText)
    resd.status = status
    resd.setData(json.dumps(resp_json).encode("utf-8"))
    return resd


def resp_usage(status, prompt_token, completion_token):
    resp_json = {"completion_tokens": completion_token, "prompt_tokens": prompt_token, "question_tokens": 4,
                 "total_tokens": prompt_token + completion_token}
    resd = ResponseData()
    resd.key = "usage"
    resd.setDataType(DataText)
    resd.status = status
    resd.setData(json.dumps(resp_json).encode("utf-8"))
    return resd


def get_param_temperature(param):
    return float(param.get('temperature', 0.95))


def get_parma_tok_k(param):
    return int(param.get('top_k', 40))


def get_param_max_new_tokens(param):
    return int(param.get('max_tokens', 2048))


def get_param_repetition_penalty(param):
    return float(param.get('repetition_penalty', 1.1))


class ThreadPool:
    def __init__(self, num_threads, wrapper):
        self.num_threads = num_threads
        self.threads = {}
        self.task_queues = {}
        self.lock = threading.Lock()
        self.wrapper = wrapper
        import logging
        self.filelogger = getFileLogger(level=logging.DEBUG)

        for i in range(num_threads):
            process_id = os.getpid()
            thread_id = "process-{}-thread-{}".format(str(process_id), str(i))
            task_queue = queue.Queue()
            self.threads[thread_id] = threading.Thread(target=self.worker, args=(thread_id, task_queue))
            self.task_queues[thread_id] = task_queue

        for thread in self.threads.values():
            thread.start()

    def worker(self, thread_id, task_queue):
        # 创建并运行事件循环
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self.task_loop(thread_id, task_queue))

    async def task_loop(self, thread_id, task_queue):
        self.filelogger.debug(f"task_loop {thread_id} enter")
        while True:
            if not task_queue.empty():
                task: PromptInferenceInfo = task_queue.get_nowait()
                if task is None:
                    break
                asyncio.create_task(task.wrapper.openaiInference(task))
                # 让出cpu
                await asyncio.sleep(0.01)
            else:
                await asyncio.sleep(0.01)
        self.filelogger.info(f"task_loop {thread_id} end")

    def alloc_min_thread(self) -> str:
        with self.lock:
            min_thread_id = min(self.threads, key=lambda thread_id: self.task_queues[thread_id].qsize())
        return min_thread_id

    def put_task(self, thread_id, task):
        with self.lock:
            self.task_queues[thread_id].put(task)

    def wait_completion(self):
        for task_queue in self.task_queues.values():
            task_queue.put(None)

        for thread in self.threads.values():
            thread.join()


# 定义服务推理逻辑
class Wrapper(WrapperBase):
    serviceId = "atp"
    version = "v1"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        import logging
        self.filelogger = getFileLogger(level=logging.DEBUG)
        self.patch_id = {}
        self.patch_id_lock = threading.Lock()
        self.request_map: dict[str, RequestInfo] = {}
        self.request_map_lock = threading.Lock()
        self.thread_pool_size = 1
        self.thread_pool = None
        self.filelogger.info(f"vllm openai client wrapper constructed")

    def get_free_port(self):
        import socket
        # 创建一个临时的socket对象
        temp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        temp_socket.bind(('localhost', 0))  # 绑定到本地地址的一个随机可用端口
        _, port = temp_socket.getsockname()  # 获取分配的端口号
        temp_socket.close()  # 关闭临时socket
        return port

    def wait_server_ready(self, server_url: str):
        import requests
        while True:
            try:
                response = requests.get(server_url + "/health", timeout=(1, 3))
                # 检查响应状态码
                if response.status_code == 200:
                    self.filelogger.info(f"{server_url} ready, {response.content}")
                    break
                else:
                    self.filelogger.info(f"{server_url} not ready")
                response.close()
            except Exception as e:
                self.filelogger.info(f"{server_url} connect exception:{e}")
            time.sleep(5)

    def setupDefaultEnv(self):
        if os.environ.get("VLLM_CONFIGURE_LOGGING", None) is None:
            os.environ['VLLM_CONFIGURE_LOGGING'] = "1"

    def setupHealthStatus(self, port):
        '''
        livenessProbe:
          exec:
            command:
              - /bin/sh
              - -c
              - "if [ -f /path/to/status ] && [ $(cat /path/to/status) -eq 0 ]; then exit 0; else exit 1; fi"
          initialDelaySeconds: 5
          periodSeconds: 10

        '''

        def run(port):
            if not os.path.exists("/var/run"):
                os.makedirs("/var/run")
            while True:
                time.sleep(3)
                with open("/var/run/wrapper_status", 'w') as h:
                    r = requests.get(f"http://localhost:{port}/health", timeout=(1, 3))
                    if r.status_code == 200:
                        h.write('0')
                    else:
                        h.write('-1')
                    h.close()

        thread = threading.Thread(target=run, args=(port,))
        thread.start()

    def wrapperInit(self, config: {}) -> int:
        self.setupDefaultEnv()
        self.base_model = os.environ.get("FULL_MODEL_PATH")
        self.pretrained_name = os.environ.get("PRETRAINED_MODEL_NAME")
        self.finetune_type = os.environ.get("FINETUNE_TYPE", 'lora')
        self.infer_type = os.environ.get("INFER_TYPE")

        self.model_generation_config: transformers.GenerationConfig = None
        if os.path.exists(os.path.join(self.base_model, "generation_config.json")):
            self.filelogger.info("Loading Default Generation Config")
            self.model_generation_config = GenerationConfig.from_pretrained(self.base_model)

        timeout = os.environ.get("READ_TIMEOUT", "30")

        self.filelogger.info(f"base_model:{self.base_model}")
        if not self.pretrained_name:
            self.filelogger.error("should have environ(PRETRAINED_MODEL_NAME,FULL_MODEL_PATH(lora weight dir))")
            return -1

        if not os.path.isdir(self.base_model):
            self.filelogger.error(f"not find the base_model in FULL_MODEL_PATH")
            return -1

        ## 获取服务监听端口
        port = self.get_free_port()
        ## 服务器实际地址
        serverUrl = f"http://127.0.0.1:{port}/v1"

        self.setupHealthStatus(port)
        ## 启动服务器进程
        launch_openai_server(self.pretrained_name, self.base_model, port, self.filelogger)
        ## 监听服务器是否启动完成
        # self.wait_server_ready(serverUrl)
        self.server_url = serverUrl
        ## 创建openai客户端
        self.client = AsyncOpenAI(base_url=serverUrl, api_key="maas", timeout=float(timeout))
        self.thread_pool = ThreadPool(num_threads=self.thread_pool_size, wrapper=self)
        self.filelogger.info(f'wrapper init success, create thread:{self.thread_pool_size}')
        return 0

    def checkValidLora(self, pretrained_name, lora_path):
        import json
        self.filelogger.debug(f'check lora {pretrained_name}, {lora_path}')
        confjson = "adapter_config.json"
        if not os.path.isdir(lora_path):
            msg = "not find  %s" % lora_path
            self.filelogger.error(msg)
            return False, msg
        files = os.listdir(lora_path)
        if not confjson in files:
            msg = "%s doesnt have file adapter_config.json" % lora_path
            self.filelogger.error(msg)
            return False, msg
        fp = open(os.path.join(lora_path, confjson), "rb")
        conf = json.load(fp)
        base_model_path = conf.get("base_model_name_or_path", "")
        if not base_model_path:
            msg = "config json not contains base_model_name_or_path...c=" % lora_path
            self.filelogger.error(msg)
            return False, msg
        self.filelogger.debug(f'base_model_path:{base_model_path}')
        if pretrained_name not in base_model_path:
            msg = f"current runntime model is {pretrained_name}, but you pass the {base_model_path}"
            self.filelogger.error(msg)
            return False, msg
        return True, "Check Success..."

    def lora_load(self):
        record_path = "/home/.atp/lora_record"
        lora_weight_path = "/home/.atp/lora_weight/"
        if not os.path.isdir(record_path):
            self.filelogger.debug("not need load lora now ")
            return
        patch_ids = os.listdir(record_path)
        paths = os.listdir(lora_weight_path)
        for patch_id in patch_ids:
            if not patch_id in paths:
                self.filelogger.warn("not find this %s" % patch_id)
                continue
            lora_weight_path_one = lora_weight_path + patch_id
            valid, msg = self.checkValidLora(self.pretrained_name, lora_weight_path_one)
            if not valid:
                self.filelogger.warn("not valid lora for this model. %s" % patch_id)
                continue
            self.patch_id_lock.acquire()
            self.patch_id[patch_id] = lora_weight_path_one
            self.patch_id_lock.release()
            self.filelogger.info(f'success wrapperLoadRes patch_id {patch_id}')

    def wrapperLoadRes(self, reqData: DataListCls, patch_id: int) -> int:
        self.patch_id_lock.acquire()
        if patch_id in self.patch_id:
            self.filelogger.warn("patch_id has exist.Please first to UnloadRes")
            self.patch_id_lock.release()
            return 0
        self.patch_id_lock.release()

        lora_weight_path = "/home/.atp/lora_weight/"
        lora_weight_path = os.path.join(lora_weight_path, str(patch_id))
        if os.path.exists(lora_weight_path):
            self.filelogger.warn("zip file has exist.Please first to UnloadRes")

        import io
        import zipfile

        byte_stream = io.BytesIO(reqData.list[0].data)
        # 解压缩 zip 文件到指定目录
        with zipfile.ZipFile(byte_stream, "r") as zip_ref:
            zip_ref.extractall(lora_weight_path)

        # write reocrd
        record_path = "/home/.atp/lora_record"
        rf = open(os.path.join(record_path, patch_id), "w")
        rf.close()

        valid, msg = self.checkValidLora(self.pretrained_name, lora_weight_path)
        if not valid:
            return -1
        self.filelogger.debug(msg)

        # lora加载
        load_resp = post_with_retry(
            url=self.server_url + "/load_lora_adapter",
            json={"lora_name": patch_id, "lora_path": lora_weight_path, "base_model_name": self.pretrained_name}
        )
        self.filelogger.info(f"patch_id:{patch_id}, load_resp:{load_resp}")

        self.patch_id_lock.acquire()
        self.patch_id[patch_id] = lora_weight_path
        self.patch_id_lock.release()
        self.filelogger.debug("Load Resource Successfully...")
        return 0

    def wrapperUnloadRes(self, presid: int) -> int:
        self.patch_id_lock.acquire()
        if presid not in self.patch_id:
            self.filelogger.error("patch_id not exist")
            self.patch_id_lock.release()
            return 0
        lora_weight_path = self.patch_id[presid]
        if not os.path.exists(lora_weight_path):
            self.filelogger.error("lora weigth path not exist")
            self.patch_id_lock.release()
            return 0

        unload_resp = post_with_retry(
            url=self.server_url + "/unload_lora_adapter",
            json={"lora_name": presid}
        )
        self.filelogger.info(f"patch_id:{presid}, unload_resp:{unload_resp}")

        import shutil
        # 递归地删除目录及其内容
        shutil.rmtree(lora_weight_path)
        del self.patch_id[presid]
        self.patch_id_lock.release()
        return 0

    def wrapperOnceExec(self, params: {}, reqData: DataListCls, usrTag: str = "", persId: int = 0) -> Response:
        prompt = get_payload_messages(reqData)
        functions = get_payload_functions(reqData)

        requestInfo = RequestInfo("", params)

        thread_id = self.thread_pool.alloc_min_thread()
        result_q = queue.Queue()
        inferenceInfo = PromptInferenceInfo(self, thread_id, RequestMode.ONCE, prompt, requestInfo, functions, result_q)
        self.filelogger.info(f'start wrapperOnceExec params:{params},request_id:{inferenceInfo.request_id}')

        self.thread_pool.put_task(thread_id, inferenceInfo)

        try:
            res = result_q.get(timeout=12)
        except queue.Empty:
            res = Response()
            content = resp_content(Once, 0, "wrapper.timeout.")
            usage = resp_usage(Once, 0, 0)
            res.list = [content, usage]

        self.filelogger.debug(
            f'success wrapperOnceExec,params:{params},prompt:{prompt},request_id:{inferenceInfo.request_id}')
        return res

    def wrapperOnceExecAsync(self, params: {}, reqData: DataListCls, usrTag: str, persId: int = 0):
        prompt = get_payload_messages(reqData)
        functions = get_payload_functions(reqData)
        if functions:
            self.filelogger.debug(f'start wrapperOnceExec functions: size {len(functions)}')
        requestInfo = RequestInfo("", params, usrTag)

        thread_id = self.thread_pool.alloc_min_thread()
        inferenceInfo = PromptInferenceInfo(self, thread_id, RequestMode.ONCE_ASYNC, prompt, requestInfo, functions)
        self.filelogger.info(f'start wrapperOnceExecAsync params:{params},request_id:{inferenceInfo.request_id}')

        self.thread_pool.put_task(thread_id, inferenceInfo)

        self.filelogger.debug(
            f'success wrapperOnceExecAsync,params:{params},prompt:{prompt},request_id:{inferenceInfo.request_id}')
        return 0

    def wrapperFini(self) -> int:
        self.thread_pool.wait_completion()
        return 0

    def wrapperError(cls, ret: int) -> str:
        if ret == 1002:
            return "Engine is busy..."
        if ret == 1001:
            return "Engine error"
        return ""

    def wrapperWrite(self, handle: str, req: DataListCls) -> int:
        self.filelogger.debug(f'start wrapperWrite handle {handle}')
        messages = req.get("messages")
        if messages:
            messages_data = messages.data
            self.request_map_lock.acquire()
            requestInfo = self.request_map[handle]
            requestInfo.requests.append(messages_data)
            self.request_map_lock.release()

            messages_status = messages.status
            if messages_status == DataEnd:
                messages_bytes = b''.join(requestInfo.requests)
                prompt = messages_bytes.decode('utf-8')
                new_messages = json.loads(prompt)
                functions = new_messages.get("functions", None)
                if functions:
                    self.filelogger.info(f'start wrapperOnceExec functions: size {len(functions)}')

                thread_id = self.thread_pool.alloc_min_thread()
                inferenceInfo = PromptInferenceInfo(self, thread_id, RequestMode.STREAM, prompt, requestInfo, functions)
                self.thread_pool.put_task(thread_id, inferenceInfo)
                self.request_map_lock.acquire()
                self.request_map[handle].requests.append(inferenceInfo.request_id)
                self.request_map_lock.release()
                self.filelogger.debug(
                    f'success wrapperWrite handle:{handle},thread_id:{thread_id},sid:{requestInfo.sid},request_id:{inferenceInfo.request_id}')
        return 0

    def wrapperCreate(self, params: {}, sid: str, persId: int = 0, usrTag: str = "") -> SessionCreateResponse:
        patch_id = str(params.get('patch_id', "0"))
        self.filelogger.info(f'start wrapperCreate {params}')
        if len(patch_id) == 0:
            patch_id = "0"
        requestInfo = RequestInfo(sid, params, usrTag)
        self.request_map_lock.acquire()
        self.request_map[requestInfo.handle] = requestInfo
        self.request_map_lock.release()

        s = SessionCreateResponse()
        s.handle = requestInfo.handle
        s.error_code = 0
        self.filelogger.debug(f'success wrapperCreate {patch_id},handle {requestInfo.handle}, sid: {sid}')
        return s

    def wrapperDestroy(self, handle: str) -> int:
        self.filelogger.debug(f'start wrapperDestroy {handle}')
        self.request_map_lock.acquire()
        requestInfo = self.request_map[handle]
        if requestInfo == None:
            self.filelogger.error("can't get this handle:" % handle)
            self.request_map_lock.release()
            return 0
        sid = requestInfo.sid
        self.request_map_lock.release()
        requestInfo.stop_q.put(True, block=False)
        self.request_map_lock.acquire()
        del self.request_map[handle]
        self.request_map_lock.release()

        self.filelogger.debug(f'success wrapperDestroy {handle} sid: {sid}')
        return 0

    def wrapperTestFunc(cls, data: [], respData: []):
        pass

    async def openaiInference(self, inferenceInfo: PromptInferenceInfo):
        requestInfo = inferenceInfo.requestInfo
        request_id = inferenceInfo.request_id
        user_tag = requestInfo.user_tag
        sid = requestInfo.sid
        is_stoped = False
        if not requestInfo.stop_q.empty():
            is_stoped = requestInfo.stop_q.get_nowait()
        if is_stoped:
            state = Once
            if inferenceInfo.mode == RequestMode.STREAM:
                state = DataEnd
            content = resp_content(state, 0, ' ')
            usage = resp_usage(state, 0, 0)
            res = Response()
            res.list = [content, usage]
            if (inferenceInfo.mode == RequestMode.STREAM or
                    inferenceInfo.mode == RequestMode.ONCE_ASYNC):
                callback(res, user_tag)
            elif inferenceInfo.mode == RequestMode.ONCE:
                inferenceInfo.result_q.put(res)
            self.filelogger.info(f'====>inference abort before infer, {requestInfo.sid}')
            return

        params = requestInfo.params
        patch_id = str(params.get('patch_id', '0'))
        if len(patch_id) == 0:
            patch_id = "0"
        prompt = inferenceInfo.prompt

        model_name = self.pretrained_name
        lora_path = None
        if self.finetune_type == 'lora' or self.finetune_type == 'qlora':
            if patch_id != '0' and patch_id != 0:
                self.patch_id_lock.acquire()
                lora_path = self.patch_id.get(patch_id)
                self.patch_id_lock.release()
                if lora_path == None:
                    self.filelogger.error(
                        f"lora or qlora infer, but lora path is invalid for patch_id:{patch_id}, will use base infer instead")
                else:
                    model_name = patch_id
        ## openai接口不支持top_k和repetition_penalty
        repetition_penalty = get_param_repetition_penalty(params)

        temperature = get_param_temperature(params)
        if IS_REASONING_MODEL and R1_FIXED_TEMPERATURE is not None:
            temperature = float(R1_FIXED_TEMPERATURE)
            self.filelogger.info(
                f"Using R1_FIXED_TEMPERATURE set in ENVIRONMENT variable: {temperature}")

        # generation_config
        if self.model_generation_config:
            top_k = self.model_generation_config.top_k
            top_p = self.model_generation_config.top_p
            do_sample = self.model_generation_config.do_sample
            temperature = self.model_generation_config.temperature
            repetition_penalty = self.model_generation_config.repetition_penalty

        max_tokens = get_param_max_new_tokens(params)

        top_k = get_parma_tok_k(params)

        # streaming case
        sparkMsgs = json.loads(prompt)["messages"]
        self.filelogger.info(f"sparkMsgs:{sparkMsgs}")
        openaiMsgs = [{"role": item["role"], "content": item["content"]} for item in sparkMsgs]

        isFc = (inferenceInfo.functions is not None and len(inferenceInfo.functions) > 0)
        try:
            full_content = ""
            reasoning_content = ""
            prompt_tokens_len = 10
            result_tokens_len = 0
            if inferenceInfo.mode != RequestMode.STREAM or isFc:
                ## 如果是functioncall请求
                if isFc:
                    openaiTools = [{"type": "function", "function": function} for function in inferenceInfo.functions]
                    resp = await self.client.chat.completions.create(
                        model=model_name,
                        messages=openaiMsgs,
                        tools=openaiTools,
                        tool_choice="auto",
                        stream=False,
                    )
                    tool_calls = []
                    if (resp is not None and
                            resp.choices[0].message.tool_calls is not None and
                            len(resp.choices[0].message.tool_calls) > 0):
                        tool_calls = [{"arguments": item.function.arguments, "name": item.function.name} for item in
                                      resp.choices[0].message.tool_calls]
                else:
                    resp = await self.client.chat.completions.create(
                        model=model_name,
                        messages=openaiMsgs,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream=False,
                    )
                    tool_calls = None
                # 整理结果
                res = Response()
                status = Once
                if inferenceInfo.mode == RequestMode.STREAM:
                    status = DataEnd
                openaiContent = resp.choices[0].message.content
                if openaiContent is None:
                    openaiContent = ' '
                content = resp_content(status, 0, openaiContent, tool_calls)
                prompt_tokens_len = resp.usage.prompt_tokens
                result_tokens_len = resp.usage.completion_tokens
                self.filelogger.info(
                    f'====>non-streaming inference end, {request_id}:-->{openaiContent}<--, toolcalls:-->{tool_calls}<--, '
                    f'sid:{sid},in_tokens:{prompt_tokens_len}, out_tokens:{result_tokens_len}')
                usage = resp_usage(status, prompt_tokens_len, result_tokens_len)
                res.list = [content, usage]
                if (inferenceInfo.mode == RequestMode.STREAM or
                        inferenceInfo.mode == RequestMode.ONCE_ASYNC):
                    callback(res, user_tag)
                elif inferenceInfo.mode == RequestMode.ONCE:
                    inferenceInfo.result_q.put(res)
                return
        except Exception as e:
            import traceback
            traceback.print_exc()
            self.filelogger.error(f"An error occurred when infer: {e}")
            return

        try:
            ## 如果是普通流式请求
            index = 0
            thinking = False
            if IS_REASONING_MODEL:
                lastMsg = openaiMsgs[-1]
                if lastMsg["role"] != "assistant":
                    openaiMsgs += [{"role": "assistant", "content": "<think>\n"}]
                    thinking = True

            streamResp = await self.client.chat.completions.create(
                model=model_name,
                messages=openaiMsgs,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=True,
                top_p=top_p,
                stream_options={
                    "include_usage": True
                }
            )
            state = DataNone

            async for chunk in streamResp:
                if not requestInfo.stop_q.empty():
                    is_stoped = requestInfo.stop_q.get_nowait()
                if is_stoped:
                    # 提前结束
                    self.filelogger.info(f'====>inference abort when infer, {requestInfo.sid}')
                    break

                if len(chunk.choices) > 0 and chunk.choices[0].delta.content is not None:
                    chunk_content = str(chunk.choices[0].delta.content)
                    res = Response()
                    if state == DataNone:
                        state = DataBegin
                    elif state == DataBegin:
                        state = DataContinue
                    if chunk_content == R1_THINK_START:
                        thinking = True
                        continue
                    if R1_THINK_END in str(chunk_content):
                        thinking = False
                        reasoning_last_chunk, answer_start_chunk = chunk_content.split(R1_THINK_END)
                        content = resp_content(state, index, "", reasoning_content=reasoning_last_chunk)
                        full_content = full_content + str(reasoning_last_chunk)
                        res.list = [content]
                        callback(res, user_tag)
                        index += 1

                        content = resp_content(state, index, answer_start_chunk, reasoning_content="")
                        full_content = full_content + str(answer_start_chunk)
                        res.list = [content]
                        callback(res, user_tag)
                        index += 1
                        continue
                    if thinking:
                        content = resp_content(state, index, "", reasoning_content=chunk_content)
                    else:
                        content = resp_content(state, index, chunk_content, reasoning_content="")

                    full_content = full_content + str(chunk_content)
                    res.list = [content]
                    callback(res, user_tag)
                    index += 1
                if chunk.usage is not None:
                    prompt_tokens_len = chunk.usage.prompt_tokens
                    result_tokens_len = chunk.usage.completion_tokens
            state = DataEnd
            content = resp_content(state, index, ' ')
            usage = resp_usage(state, prompt_tokens_len, result_tokens_len)
            res = Response()
            res.list = [content, usage]
            callback(res, user_tag)

        except openai.APITimeoutError as te:
            self.filelogger.error(f"Connect Engine Timeout: {te}")
            res = Response()
            res = res.response_err(1002)
            callback(res, user_tag)
        except httpx.TimeoutException as te:
            self.filelogger.error(f"Connect Engine Timeout: {te}")
            res = Response()
            res = res.response_err(1003)
            callback(res, user_tag)
        except Exception as e:
            import traceback
            traceback.print_exc()
            self.filelogger.error(f"An error occurred when infer: {e}, ret 1001")
            res = Response()
            res = res.response_err(1001)
            callback(res, user_tag)

        self.filelogger.info(
            f'====>streaming inference end, {requestInfo.sid}:{full_content}, sid:{sid},in_tokens:{prompt_tokens_len}, out_tokens:{result_tokens_len}')
        return


def generate_template(t="alpaca"):
    template = "/home/aiges/default.jinja"
    with open(template, 'w') as f:
        if t == "alpaca":
            f.write('''{{ (messages|selectattr('role', 'equalto', 'system')|list|last).content|trim if (messages|selectattr('role', 'equalto', 'system')|list) else '' }}

{% for message in messages %}
{% if message['role'] == 'user' %}
### Instruction:
{{ message['content']|trim -}}
{% if not loop.last %}


{% endif %}
{% elif message['role'] == 'assistant' %}
### Response:
{{ message['content']|trim -}}
{% if not loop.last %}


{% endif %}
{% elif message['role'] == 'user_context' %}
### Input:
{{ message['content']|trim -}}
{% if not loop.last %}


{% endif %}
{% endif %}
{% endfor %}
{% if add_generation_prompt and messages[-1]['role'] != 'assistant' %}
### Response:
{% endif %}''')
        elif t == "llama2_json":
            f.write('''{{- bos_token }}
{%- if custom_tools is defined %}
    {%- set tools = custom_tools %}
{%- endif %}
{%- if not tools_in_user_message is defined %}
    {%- set tools_in_user_message = false %}
{%- endif %}
{%- if not date_string is defined %}
    {%- if strftime_now is defined %}
        {%- set date_string = strftime_now("%d %b %Y") %}
    {%- else %}
        {%- set date_string = "26 Jul 2024" %}
    {%- endif %}
{%- endif %}
{%- if not tools is defined %}
    {%- set tools = none %}
{%- endif %}

{#- Find out if there are any images #}
{% set image_ns = namespace(has_images=false) %}
{%- for message in messages %}
    {%- for content in message['content'] %}
        {%- if content['type'] == 'image' %}
            {%- set image_ns.has_images = true %}
        {%- endif %}
    {%- endfor %}
{%- endfor %}

{#- This block extracts the system message, so we can slot it into the right place. #}
{%- if messages[0]['role'] == 'system' %}
    {%- if messages[0]['content'] is string %}
        {%- set system_message = messages[0]['content']|trim %}
    {%- else %}
        {%- set system_message = messages[0]['content'][0]['text']|trim %}
    {%- endif %}
    {%- set messages = messages[1:] %}
{%- else %}
    {%- if tools is not none %}
        {%- set system_message = "You are a helpful assistant with tool calling capabilities. Only reply with a tool call if the function exists in the library provided by the user. If it doesn't exist, just reply directly in natural language. When you receive a tool call response, use the output to format an answer to the original user question." %}
    {%- else %}
        {%- set system_message = "" %}
    {%- endif %}
{%- endif %}

{#- System message if there are no images, if the user supplied one, or if tools are used (default tool system message) #}
{%- if system_message or not image_ns.has_images %}
    {{- "<|start_header_id|>system<|end_header_id|>\n\n" }}
    {%- if tools is not none %}
        {{- "Environment: ipython\n" }}
    {%- endif %}
    {{- "Cutting Knowledge Date: December 2023\n" }}
    {{- "Today Date: " + date_string + "\n\n" }}
    {%- if tools is not none and not tools_in_user_message %}
        {{- "You have access to the following functions. To call a function, please respond with JSON for a function call. " }}
        {{- 'Respond in the format {"name": function name, "parameters": dictionary of argument name and its value}. ' }}
        {{- "Do not use variables.\n\n" }}
        {%- for t in tools %}
            {{- t | tojson(indent=4) }}
            {{- "\n\n" }}
        {%- endfor %}
    {%- endif %}
    {{- system_message }}
    {{- "<|eot_id|>" }}
{%- endif %}

{#- Custom tools are passed in a user message with some extra guidance #}
{%- if tools_in_user_message and not tools is none %}
    {#- Extract the first user message so we can plug it in here #}
    {%- if messages | length != 0 %}
        {%- if messages[0]['content'] is string %}
            {%- set first_user_message = messages[0]['content']|trim %}
        {%- else %}
            {%- set first_user_message = messages[0]['content'] | selectattr('type', 'equalto', 'text') | map(attribute='text') | map('trim') | join('\n') %}
        {%- endif %}
        {%- set messages = messages[1:] %}
    {%- else %}
        {{- raise_exception("Cannot put tools in the first user message when there's no first user message!") }}
    {%- endif %}
    {{- '<|start_header_id|>user<|end_header_id|>\n\n' -}}
    {{- "Given the following functions, please respond with a JSON for a function call " }}
    {{- "with its proper arguments that best answers the given prompt.\n\n" }}
    {{- 'Respond in the format {"name": function name, "parameters": dictionary of argument name and its value}. ' }}
    {{- "Do not use variables.\n\n" }}
    {%- for t in tools %}
        {{- t | tojson(indent=4) }}
        {{- "\n\n" }}
    {%- endfor %}
    {{- first_user_message + "<|eot_id|>"}}
{%- endif %}

{%- for message in messages %}
    {%- if not (message.role == 'ipython' or message.role == 'tool' or 'tool_calls' in message) %}
        {{- '<|start_header_id|>' + message['role'] + '<|end_header_id|>\n\n' }}
        {%- if message['content'] is string %}
            {{- message['content'] | trim}}
        {%- else %}
            {%- for content in message['content'] %}
                {%- if content['type'] == 'image' %}
                    {{- '<|image|>' }}
                {%- elif content['type'] == 'text' %}
                    {{- content['text'] | trim }}
                {%- endif %}
            {%- endfor %}
        {%- endif %}
        {{- '<|eot_id|>' }}
    {%- elif 'tool_calls' in message %}
        {%- if not message.tool_calls|length == 1 %}
            {{- raise_exception("This model only supports single tool-calls at once!") }}
        {%- endif %}
        {%- set tool_call = message.tool_calls[0].function %}
        {{- '<|start_header_id|>assistant<|end_header_id|>\n\n' -}}
        {{- '{"name": "' + tool_call.name + '", ' }}
        {{- '"parameters": ' }}
        {{- tool_call.arguments | tojson }}
        {{- "}" }}
        {{- "<|eot_id|>" }}
    {%- elif message.role == "tool" or message.role == "ipython" %}
        {{- "<|start_header_id|>ipython<|end_header_id|>\n\n" }}
        {%- if message.content is string %}
            {{- { "output": message.content } | tojson }}
        {%- else %}
            {%- for content in message['content']  %}
                {%- if content['type']  == 'text' %}
                    {{- { "output": content['text']  } | tojson }}
                {%- endif %}
            {%- endfor %}
        {%- endif %}
        {{- "<|eot_id|>" }}
    {%- endif %}
{%- endfor %}
{%- if add_generation_prompt %}
    {{- '<|start_header_id|>assistant<|end_header_id|>\n\n' }}
{%- endif %}''')

    return template
